@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+require 'pry/helpers/documentation_helpers'
+
 class Pry
   class << self
     # If the given object is a `Pry::Method`, return it unaltered. If it's
@@ -12,8 +14,17 @@ class Pry
     end
   end
 
+  # This class wraps the normal `Method` and `UnboundMethod` classes
+  # to provide extra functionality useful to Pry.
   class Method
-    include RbxMethod if Helpers::BaseHelpers.rbx?
+    require 'pry/method/weird_method_locator'
+    require 'pry/method/disowned'
+
+    extend Helpers::BaseHelpers
+    include Helpers::BaseHelpers
+    include RbxMethod if rbx?
+    include Helpers::DocumentationHelpers
+    include CodeObject::Helpers
 
     class << self
       # Given a string representing a method name and optionally a binding to
@@ -35,18 +46,21 @@ class Pry
           from_binding(target)
         elsif name.to_s =~ /(.+)\#(\S+)\Z/
           context, meth_name = $1, $2
-          from_module(target.eval(context), meth_name)
-        elsif name.to_s =~ /(.+)\.(\S+)\Z/
-          context, meth_name = $1, $2
-          from_obj(target.eval(context), meth_name)
+          from_module(target.eval(context), meth_name, target)
+        elsif name.to_s =~ /(.+)(\.|::)(\S+)\Z/
+          context, meth_name = $1, $3
+          from_obj(target.eval(context), meth_name, target)
         elsif options[:instance]
-          from_module(target.eval("self"), name)
+          from_module(target.eval("self"), name, target)
         elsif options[:methods]
-          from_obj(target.eval("self"), name)
+          from_obj(target.eval("self"), name, target)
         else
           from_str(name, target, :instance => true) or
             from_str(name, target, :methods => true)
         end
+
+      rescue Pry::RescuableException
+        nil
       end
 
       # Given a `Binding`, try to extract the `::Method` it originated from and
@@ -57,59 +71,54 @@ class Pry
       # @return [Pry::Method, nil]
       #
       def from_binding(b)
-        meth_name = b.eval('__method__')
-        if [:__script__, nil, :__binding__, :__binding_impl__].include?(meth_name)
+        meth_name = b.eval('::Kernel.__method__')
+        if [:__script__, nil].include?(meth_name)
           nil
         else
           method = begin
-                     new(b.eval("method(#{meth_name.to_s.inspect})"))
+                     if Object === b.eval('self')
+                       new(Kernel.instance_method(:method).bind(b.eval("self")).call(meth_name))
+                     else
+                       new(b.eval('class << self; self; end.instance_method(::Kernel.__method__).bind(self)'))
+                     end
                    rescue NameError, NoMethodError
                      Disowned.new(b.eval('self'), meth_name.to_s)
                    end
 
-          # it's possible in some cases that the method we find by this approach is a sub-method of
-          # the one we're currently in, consider:
-          #
-          # class A; def b; binding.pry; end; end
-          # class B < A; def b; super; end; end
-          #
-          # Given that we can normally find the source_range of methods, and that we know which
-          # __FILE__ and __LINE__ the binding is at, we can hope to disambiguate these cases.
-          #
-          # This obviously won't work if the source is unavaiable for some reason, or if both
-          # methods have the same __FILE__ and __LINE__, or if we're in rbx where b.eval('__LINE__')
-          # is broken.
-          #
-          guess = method
-
-          while guess
-            # needs rescue if this is a Disowned method or a C method or something...
-            # TODO: Fix up the exception handling so we don't need a bare rescue
-            if (guess.source_file && guess.source_range rescue false) &&
-                File.expand_path(guess.source_file) == File.expand_path(b.eval('__FILE__')) &&
-                guess.source_range.include?(b.eval('__LINE__'))
-              return guess
-            else
-              guess = guess.super
-            end
+          if WeirdMethodLocator.weird_method?(method, b)
+            WeirdMethodLocator.new(method, b).get_method || method
+          else
+            method
           end
-
-          # Uhoh... none of the methods in the chain had the right __FILE__ and __LINE__
-          # This may be caused by rbx https://github.com/rubinius/rubinius/issues/953,
-          # or other unknown circumstances (TODO: we should warn the user when this happens)
-          method
         end
       end
 
+      # In order to support 2.0 Refinements we need to look up methods
+      # inside the relevant Binding.
+      # @param [Object] obj The owner/receiver of the method.
+      # @param [Symbol] method_name The name of the method.
+      # @param [Symbol] method_type The type of method: :method or :instance_method
+      # @param [Binding] target The binding where the method is looked up.
+      # @return [Method, UnboundMethod] The 'refined' method object.
+      def lookup_method_via_binding(obj, method_name, method_type, target=TOPLEVEL_BINDING)
+        Pry.current[:obj] = obj
+        Pry.current[:name] = method_name
+        receiver = obj.is_a?(Module) ? "Module" : "Kernel"
+        target.eval("::#{receiver}.instance_method(:#{method_type}).bind(Pry.current[:obj]).call(Pry.current[:name])")
+      ensure
+        Pry.current[:obj] = Pry.current[:name] = nil
+      end
+
       # Given a `Class` or `Module` and the name of a method, try to
       # instantiate a `Pry::Method` containing the instance method of
       # that name. Return `nil` if no such method exists.
       #
       # @param [Class, Module] klass
       # @param [String] name
+      # @param [Binding] target The binding where the method is looked up.
       # @return [Pry::Method, nil]
-      def from_class(klass, name)
-        new(klass.instance_method(name)) rescue nil
+      def from_class(klass, name, target=TOPLEVEL_BINDING)
+        new(lookup_method_via_binding(klass, name, :instance_method, target)) rescue nil
       end
       alias from_module from_class
 
@@ -119,23 +128,26 @@ class Pry
       #
       # @param [Object] obj
       # @param [String] name
+      # @param [Binding] target The binding where the method is looked up.
       # @return [Pry::Method, nil]
-      def from_obj(obj, name)
-        new(obj.method(name)) rescue nil
+      def from_obj(obj, name, target=TOPLEVEL_BINDING)
+        new(lookup_method_via_binding(obj, name, :method, target)) rescue nil
       end
 
       # Get all of the instance methods of a `Class` or `Module`
       # @param [Class,Module] klass
+      # @param [Boolean] include_super Whether to include methods from ancestors.
       # @return [Array[Pry::Method]]
-      def all_from_class(klass)
-        all_from_common(klass, :instance_method)
+      def all_from_class(klass, include_super=true)
+        all_from_common(klass, :instance_method, include_super)
       end
 
       # Get all of the methods on an `Object`
       # @param [Object] obj
+      # @param [Boolean] include_super Whether to include methods from ancestors.
       # @return [Array[Pry::Method]]
-      def all_from_obj(obj)
-        all_from_common(obj, :method)
+      def all_from_obj(obj, include_super=true)
+        all_from_common(obj, :method, include_super)
       end
 
       # Get every `Class` and `Module`, in order, that will be checked when looking
@@ -161,6 +173,19 @@ class Pry
         ([klass] + klass.ancestors).uniq
       end
 
+      def method_definition?(name, definition_line)
+        singleton_method_definition?(name, definition_line) ||
+          instance_method_definition?(name, definition_line)
+      end
+
+      def singleton_method_definition?(name, definition_line)
+        /^define_singleton_method\(?\s*[:\"\']#{name}|^def\s*self\.#{name}/ =~ definition_line.strip
+      end
+
+      def instance_method_definition?(name, definition_line)
+        /^define_method\(?\s*[:\"\']#{name}|^def\s*#{name}/ =~ definition_line.strip
+      end
+
       private
 
       # See all_from_class and all_from_obj.
@@ -168,22 +193,14 @@ class Pry
       # If method_type is :method, obj can be any `Object`
       #
       # N.B. we pre-cache the visibility here to avoid O(N²) behaviour in "ls".
-      def all_from_common(obj, method_type)
+      def all_from_common(obj, method_type, include_super=true)
         %w(public protected private).map do |visibility|
-          safe_send(obj, :"#{visibility}_#{method_type}s").map do |method_name|
+          safe_send(obj, :"#{visibility}_#{method_type}s", include_super).map do |method_name|
             new(safe_send(obj, method_type, method_name), :visibility => visibility.to_sym)
           end
         end.flatten(1)
       end
 
-      # Acts like send but ignores any methods defined below Object or Class in the
-      # inheritance hierarchy.
-      # This is required to introspect methods on objects like Net::HTTP::Get that
-      # have overridden the `method` method.
-      def safe_send(obj, method, *args, &block)
-        (Module === obj ? Module : Object).instance_method(method).bind(obj).call(*args, &block)
-      end
-
       # Get the singleton classes of superclasses that could define methods on
       # the given class object, and any modules they include.
       # If a module is included at multiple points in the ancestry, only
@@ -202,7 +219,7 @@ class Pry
     # A new instance of `Pry::Method` wrapping the given `::Method`, `UnboundMethod`, or `Proc`.
     #
     # @param [::Method, UnboundMethod, Proc] method
-    # @param [Hash] known_info, can be used to pre-cache expensive to compute stuff.
+    # @param [Hash] known_info Can be used to pre-cache expensive to compute stuff.
     # @return [Pry::Method]
     def initialize(method, known_info={})
       @method = method
@@ -221,6 +238,12 @@ class Pry
       @wrapped_owner ||= Pry::WrappedModule.new(owner)
     end
 
+    # Get underlying object wrapped by this Pry::Method instance
+    # @return [Method, UnboundMethod, Proc]
+    def wrapped
+      @method
+    end
+
     # Is the method undefined? (aka `Disowned`)
     # @return [Boolean] false
     def undefined?
@@ -239,21 +262,33 @@ class Pry
     # @return [String, nil] The source code of the method, or `nil` if it's unavailable.
     def source
       @source ||= case source_type
-        when :c
-          info = pry_doc_info
-          if info and info.source
-            code = strip_comments_from_c_code(info.source)
-          end
-        when :ruby
-          if Helpers::BaseHelpers.rbx? && core?
-            code = core_code
-          elsif pry_method?
-            code = Pry.new(:input => StringIO.new(Pry.line_buffer[source_line..-1].join), :prompt => proc {""}, :hooks => Pry::Hooks.new).r
-          else
-            code = @method.source
-          end
-          strip_leading_whitespace(code)
-        end
+                  when :c
+                    info = pry_doc_info
+                    if info and info.source
+                      code = strip_comments_from_c_code(info.source)
+                    end
+                  when :ruby
+                    # clone of MethodSource.source_helper that knows to use our
+                    # hacked version of source_location for rbx core methods, and
+                    # our input buffer for methods defined in (pry)
+                    file, line = *source_location
+                    raise SourceNotFoundError, "Could not locate source for #{name_with_owner}!" unless file
+
+                    begin
+                      code = Pry::Code.from_file(file).expression_at(line)
+                    rescue SyntaxError => e
+                      raise MethodSource::SourceNotFoundError.new(e.message)
+                    end
+                    strip_leading_whitespace(code)
+                  end
+    end
+
+    # Can we get the source code for this method?
+    # @return [Boolean]
+    def source?
+      !!source
+    rescue MethodSource::SourceNotFoundError
+      false
     end
 
     # @return [String, nil] The documentation for the method, or `nil` if it's
@@ -265,12 +300,12 @@ class Pry
           info = pry_doc_info
           info.docstring if info
         when :ruby
-          if Helpers::BaseHelpers.rbx? && core?
-            strip_leading_hash_and_whitespace_from_ruby_comments(core_doc)
+          if rbx? && !pry_method?
+            get_comment_content(core_doc)
           elsif pry_method?
-            raise CommandError, "Can't view doc for a REPL-defined method."
+            get_comment_content(doc_for_pry_method)
           else
-            strip_leading_hash_and_whitespace_from_ruby_comments(@method.comment)
+            get_comment_content(@method.comment)
           end
         end
     end
@@ -281,11 +316,20 @@ class Pry
       source_location.nil? ? :c : :ruby
     end
 
+    def source_location
+      if @method.source_location && rbx?
+        file, line = @method.source_location
+        [RbxPath.convert_path_to_full(file), line]
+      else
+        @method.source_location
+      end
+    end
+
     # @return [String, nil] The name of the file the method is defined in, or
     #   `nil` if the filename is unavailable.
     def source_file
       if source_location.nil?
-        if !Helpers::BaseHelpers.rbx? and source_type == :c
+        if !rbx? and source_type == :c
           info = pry_doc_info
           info.file if info
         end
@@ -303,7 +347,7 @@ class Pry
     # @return [Range, nil] The range of lines in `source_file` which contain
     #    the method's definition, or `nil` if that information is unavailable.
     def source_range
-      source_location.nil? ? nil : (source_line)...(source_line + source.lines.count)
+      source_location.nil? ? nil : (source_line)..(source_line + source.lines.count - 1)
     end
 
     # @return [Symbol] The visibility of the method. May be `:public`,
@@ -369,11 +413,44 @@ class Pry
       !!(source_file and source_file =~ /(\(.*\))|<.*>/)
     end
 
+    # @return [Boolean] Whether the method is unbound.
+    def unbound_method?
+      is_a?(::UnboundMethod)
+    end
+
+    # @return [Boolean] Whether the method is bound.
+    def bound_method?
+      is_a?(::Method)
+    end
+
+    # @return [Boolean] Whether the method is a singleton method.
+    def singleton_method?
+      wrapped_owner.singleton_class?
+    end
+
     # @return [Boolean] Was the method defined within the Pry REPL?
     def pry_method?
       source_file == Pry.eval_path
     end
 
+    # @return [Array<String>] All known aliases for the method.
+    # @note On Ruby 1.8 this method always returns an empty Array for methods
+    #   implemented in C.
+    def aliases
+      owner = @method.owner
+      # Avoid using `to_sym` on {Method#name}, which returns a `String`, because
+      # it won't be garbage collected.
+      name = @method.name
+
+      alias_list = owner.instance_methods.combination(2).select do |pair|
+        pair.include?(name) &&
+          owner.instance_method(pair.first) == owner.instance_method(pair.last)
+      end.flatten
+      alias_list.delete(name)
+
+      alias_list.map(&:to_s)
+    end
+
     # @return [Boolean] Is the method definitely an alias?
     def alias?
       name != original_name
@@ -382,7 +459,7 @@ class Pry
     # @return [Boolean]
     def ==(obj)
       if obj.is_a? Pry::Method
-        super
+        obj == @method
       else
         @method == obj
       end
@@ -407,109 +484,68 @@ class Pry
     end
 
     private
-      # @return [YARD::CodeObjects::MethodObject]
-      # @raise [CommandError] Raises when the method can't be found or `pry-doc` isn't installed.
-      def pry_doc_info
-        if Pry.config.has_pry_doc
-          Pry::MethodInfo.info_for(@method) or raise CommandError, "Cannot locate this method: #{name}."
+
+    # @return [YARD::CodeObjects::MethodObject]
+    # @raise [CommandError] when the method can't be found or `pry-doc` isn't installed.
+    def pry_doc_info
+      if Pry.config.has_pry_doc
+        Pry::MethodInfo.info_for(@method) or raise CommandError, "Cannot locate this method: #{name}. (source_location returns nil)"
+      else
+        fail_msg = "Cannot locate this method: #{name}."
+        if mri_18? || mri_19?
+          fail_msg += ' Try `gem-install pry-doc` to get access to Ruby Core documentation.'
+        end
+        raise CommandError, fail_msg
+      end
+    end
+
+    # FIXME: a very similar method to this exists on WrappedModule: extract_doc_for_candidate
+    def doc_for_pry_method
+      _, line_num = source_location
+
+      buffer = ""
+      Pry.line_buffer[0..(line_num - 1)].each do |line|
+        # Add any line that is a valid ruby comment,
+        # but clear as soon as we hit a non comment line.
+        if (line =~ /^\s*#/) || (line =~ /^\s*$/)
+          buffer << line.lstrip
         else
-          raise CommandError, "Cannot locate this method: #{name}. Try `gem install pry-doc` to get access to Ruby Core documentation."
+          buffer.replace("")
         end
       end
 
-      # @param [String] code
-      # @return [String]
-      def strip_comments_from_c_code(code)
-        code.sub(/\A\s*\/\*.*?\*\/\s*/m, '')
-      end
+      buffer
+    end
 
-      # @param [String] comment
-      # @return [String]
-      def strip_leading_hash_and_whitespace_from_ruby_comments(comment)
-        comment = comment.dup
-        comment.gsub!(/\A\#+?$/, '')
-        comment.gsub!(/^\s*#/, '')
-        strip_leading_whitespace(comment)
-      end
-
-      # @param [String] text
-      # @return [String]
-      def strip_leading_whitespace(text)
-        Pry::Helpers::CommandHelpers.unindent(text)
-      end
-
-      # @param [Class,Module] the ancestors to investigate
-      # @return [Method] the unwrapped super-method
-      def super_using_ancestors(ancestors, times=1)
-        next_owner = self.owner
-        times.times do
-          i = ancestors.index(next_owner) + 1
-          while ancestors[i] && !(ancestors[i].method_defined?(name) || ancestors[i].private_method_defined?(name))
-            i += 1
-          end
-          next_owner = ancestors[i] or return nil
+    # @param [Class, Module] ancestors The ancestors to investigate
+    # @return [Method] The unwrapped super-method
+    def super_using_ancestors(ancestors, times=1)
+      next_owner = self.owner
+      times.times do
+        i = ancestors.index(next_owner) + 1
+        while ancestors[i] && !(ancestors[i].method_defined?(name) || ancestors[i].private_method_defined?(name))
+          i += 1
         end
-        next_owner.instance_method(name) rescue nil
+        next_owner = ancestors[i] or return nil
       end
 
-      # @param [String] first_ln The first line of a method definition.
-      # @return [String, nil]
-      def method_name_from_first_line(first_ln)
-        return nil if first_ln.strip !~ /^def /
+      safe_send(next_owner, :instance_method, name) rescue nil
+    end
 
-        tokens = CodeRay.scan(first_ln, :ruby)
-        tokens = tokens.tokens.each_slice(2) if tokens.respond_to?(:tokens)
-        tokens.each_cons(2) do |t1, t2|
-          if t2.last == :method || t2.last == :ident && t1 == [".", :operator]
-            return t2.first
-          end
+    # @param [String] first_ln The first line of a method definition.
+    # @return [String, nil]
+    def method_name_from_first_line(first_ln)
+      return nil if first_ln.strip !~ /^def /
+
+      tokens = CodeRay.scan(first_ln, :ruby)
+      tokens = tokens.tokens.each_slice(2) if tokens.respond_to?(:tokens)
+      tokens.each_cons(2) do |t1, t2|
+        if t2.last == :method || t2.last == :ident && t1 == [".", :operator]
+          return t2.first
         end
-
-        nil
       end
 
-    # A Disowned Method is one that's been removed from the class on which it was defined.
-    #
-    # e.g.
-    # class C
-    #   def foo
-    #     C.send(:undefine_method, :foo)
-    #     Pry::Method.from_binding(binding)
-    #   end
-    # end
-    #
-    # In this case we assume that the "owner" is the singleton class of the receiver.
-    #
-    # This occurs mainly in Sinatra applications.
-    class Disowned < Method
-      attr_reader :receiver, :name
-
-      # Create a new Disowned method.
-      #
-      # @param [Object] receiver
-      # @param [String] method_name
-      def initialize(*args)
-        @receiver, @name = *args
-      end
-
-      # Is the method undefined? (aka `Disowned`)
-      # @return [Boolean] true
-      def undefined?
-        true
-      end
-
-      # Get the hypothesized owner of the method.
-      #
-      # @return [Object]
-      def owner
-        class << receiver; self; end
-      end
-
-      # Raise a more useful error message instead of trying to forward to nil.
-      def method_missing(meth_name, *args, &block)
-        raise "Cannot call '#{meth_name}' on an undef'd method." if method(:name).respond_to?(meth_name)
-        Object.instance_method(:method_missing).bind(self).call(meth_name, *args, &block)
-      end
+      nil
     end
   end
 end
