@@ -5,7 +5,13 @@ require 'pry/config'
 class Pry
 
   # The RC Files to load.
-  RC_FILES = ["~/.pryrc", "./.pryrc"]
+  HOME_RC_FILE = "~/.pryrc"
+  LOCAL_RC_FILE = "./.pryrc"
+
+  # @return [Hash] Pry's `Thread.current` hash
+  def self.current
+    Thread.current[:__pry__] ||= {}
+  end
 
   # class accessors
   class << self
@@ -43,26 +49,44 @@ class Pry
     # @return [Boolean] Whether Pry was activated from the command line.
     attr_accessor :cli
 
-    # @return [Fixnum] The number of active Pry sessions.
-    attr_accessor :active_sessions
+    # @return [Boolean] Whether Pry sessions are quiet by default.
+    attr_accessor :quiet
+
+    # @return [Binding] A top level binding with no local variables
+    attr_accessor :toplevel_binding
+
+    # @return [Exception, nil] The last pry internal error.
+    #   (a CommandError in most cases)
+    attr_accessor :last_internal_error
 
     # plugin forwardables
     def_delegators :@plugin_manager, :plugins, :load_plugins, :locate_plugins
 
     delegate_accessors :@config, :input, :output, :commands, :prompt, :print, :exception_handler,
-      :hooks, :color, :pager, :editor, :memory_size, :input_stack
+      :hooks, :color, :pager, :editor, :memory_size, :input_stack, :extra_sticky_locals
+  end
+
+  # Load the given file in the context of `Pry.toplevel_binding`
+  # @param [String] file_name The unexpanded file path.
+  def self.load_file_at_toplevel(file_name)
+    full_name = File.expand_path(file_name)
+    begin
+      toplevel_binding.eval(File.read(full_name), full_name) if File.exists?(full_name)
+    rescue RescuableException => e
+      puts "Error loading #{file_name}: #{e}\n#{e.backtrace.first}"
+    end
   end
 
   # Load the rc files given in the `Pry::RC_FILES` array.
   # This method can also be used to reload the files if they have changed.
   def self.load_rc
-    files = RC_FILES.collect { |file_name| File.expand_path(file_name) }.uniq
-    files.each do |file_name|
-      begin
-        load(file_name) if File.exists?(file_name)
-      rescue RescuableException => e
-        puts "Error loading #{file_name}: #{e}"
-      end
+    load_file_at_toplevel(HOME_RC_FILE)
+  end
+
+  # Load the local RC file (./.pryrc)
+  def self.load_local_rc
+    unless File.expand_path(HOME_RC_FILE) == File.expand_path(LOCAL_RC_FILE)
+      load_file_at_toplevel(LOCAL_RC_FILE)
     end
   end
 
@@ -83,12 +107,12 @@ class Pry
   # Including: loading .pryrc, loading plugins, loading requires, and
   # loading history.
   def self.initial_session_setup
-
-    return if !initial_session?
+    return unless initial_session?
 
     # note these have to be loaded here rather than in pry_instance as
     # we only want them loaded once per entire Pry lifetime.
     load_rc if Pry.config.should_load_rc
+    load_local_rc if Pry.config.should_load_local_rc
     load_plugins if Pry.config.should_load_plugins
     load_requires if Pry.config.should_load_requires
     load_history if Pry.config.history.should_load
@@ -98,15 +122,23 @@ class Pry
   end
 
   # Start a Pry REPL.
-  # This method also loads the files specified in `Pry::RC_FILES` the
+  # This method also loads the ~/.pryrc and ./.pryrc as necessary
   # first time it is invoked.
   # @param [Object, Binding] target The receiver of the Pry session
   # @param [Hash] options
   # @option options (see Pry#initialize)
   # @example
   #   Pry.start(Object.new, :input => MyInput.new)
-  def self.start(target=TOPLEVEL_BINDING, options={})
-    target = Pry.binding_for(target)
+  def self.start(target=nil, options={})
+    return if ENV['DISABLE_PRY']
+
+    if in_critical_section?
+      output.puts "ERROR: Pry started inside Pry."
+      output.puts "This can happen if you have a binding.pry inside a #to_s or #inspect function."
+      return
+    end
+
+    target = Pry.binding_for(target || toplevel_binding)
     initial_session_setup
 
     # create the Pry instance to manage the session
@@ -119,12 +151,7 @@ class Pry
     pry_instance.backtrace.shift if pry_instance.backtrace.first =~ /pry.*core_extensions.*pry/
 
     # yield the binding_stack to the hook for modification
-    pry_instance.exec_hook(
-                           :when_started,
-                           target,
-                           options,
-                           pry_instance
-                           )
+    pry_instance.exec_hook(:when_started, target, options, pry_instance)
 
     if !pry_instance.binding_stack.empty?
       head = pry_instance.binding_stack.pop
@@ -132,8 +159,24 @@ class Pry
       head = target
     end
 
+    # Clear the line before starting Pry. This fixes the issue discussed here:
+    # https://github.com/pry/pry/issues/566
+    if Pry.config.auto_indent
+      Kernel.print Pry::Helpers::BaseHelpers.windows_ansi? ? "\e[0F" : "\e[0G"
+    end
+
     # Enter the matrix
     pry_instance.repl(head)
+  rescue Pry::TooSafeException
+    puts "ERROR: Pry cannot work with $SAFE > 0"
+    raise
+  end
+
+  # Execute the file through the REPL loop, non-interactively.
+  # @param [String] file_name File name to load through the REPL.
+  def self.load_file_through_repl(file_name)
+    require "pry/repl_file_loader"
+    REPLFileLoader.new(file_name).load
   end
 
   # An inspector that clips the output to `max_length` chars.
@@ -175,7 +218,7 @@ class Pry
 
   # Run a Pry command from outside a session. The commands available are
   # those referenced by `Pry.commands` (the default command set).
-  # @param [String] arg_string The Pry command (including arguments,
+  # @param [String] command_string The Pry command (including arguments,
   #   if any).
   # @param [Hash] options Optional named parameters.
   # @return [Object] The return value of the Pry command.
@@ -199,7 +242,9 @@ class Pry
 
     output = options[:show_output] ? options[:output] : StringIO.new
 
-    Pry.new(:output => output, :input => StringIO.new(command_string), :commands => options[:commands], :prompt => proc {""}, :hooks => Pry::Hooks.new).rep(options[:context])
+    Pry.new(:output => output, :input => StringIO.new("#{command_string}\nexit-all\n"),
+            :commands => options[:commands],
+            :prompt => proc {""}, :hooks => Pry::Hooks.new).repl(options[:context])
   end
 
   def self.default_editor_for_platform
@@ -215,10 +260,36 @@ class Pry
     end
   end
 
+  def self.auto_resize!
+    Pry.require_readline
+    ver = Readline::VERSION
+    if ver[/edit/i]
+      warn <<-EOT
+Readline version #{ver} detected - will not auto_resize! correctly.
+  For the fix, use GNU Readline instead:
+  https://github.com/guard/guard/wiki/Add-proper-Readline-support-to-Ruby-on-Mac-OS-X
+      EOT
+      return
+    end
+    trap :WINCH do
+      begin
+        Readline.set_screen_size *Terminal.size!
+      rescue => e
+        warn "\nPry.auto_resize!'s Readline.set_screen_size failed: #{e}"
+      end
+      begin
+        Readline.refresh_line
+      rescue => e
+        warn "\nPry.auto_resize!'s Readline.refresh_line failed: #{e}"
+      end
+    end
+  end
+
   def self.set_config_defaults
-    config.input = Readline
+    config.input = proc { Pry.require_readline; Readline }
     config.output = $stdout
     config.commands = Pry::Commands
+    config.prompt_name = DEFAULT_PROMPT_NAME
     config.prompt = DEFAULT_PROMPT
     config.print = DEFAULT_PRINT
     config.exception_handler = DEFAULT_EXCEPTION_HANDLER
@@ -231,12 +302,15 @@ class Pry
     config.system = DEFAULT_SYSTEM
     config.editor = default_editor_for_platform
     config.should_load_rc = true
+    config.should_load_local_rc = true
     config.should_trap_interrupts = Helpers::BaseHelpers.jruby?
     config.disable_auto_reload = false
     config.command_prefix = ""
-    config.auto_indent = true
+    config.auto_indent = Helpers::BaseHelpers.use_ansi_codes?
     config.correct_indent = true
     config.collision_warning = false
+    config.output_prefix = "=> "
+    config.completer = Pry::BondCompleter
 
     config.gist ||= OpenStruct.new
     config.gist.inspecter = proc(&:pretty_inspect)
@@ -249,22 +323,30 @@ class Pry
     config.history ||= OpenStruct.new
     config.history.should_save = true
     config.history.should_load = true
-    config.history.file = File.expand_path("~/.pry_history")
+    config.history.file = File.expand_path("~/.pry_history") rescue nil
+
+    if config.history.file.nil?
+      config.should_load_rc = false
+      config.history.should_save = false
+      config.history.should_load = false
+    end
 
     config.control_d_handler = DEFAULT_CONTROL_D_HANDLER
 
     config.memory_size = 100
 
+    config.extra_sticky_locals = {}
+
     config.ls ||= OpenStruct.new({
-      :heading_color            => :default,
+      :heading_color            => :bright_blue,
 
       :public_method_color      => :default,
-      :private_method_color     => :green,
-      :protected_method_color   => :yellow,
+      :private_method_color     => :blue,
+      :protected_method_color   => :blue,
       :method_missing_color     => :bright_red,
 
-      :local_var_color          => :default,
-      :pry_var_color            => :red,         # e.g. _, _pry_, _file_
+      :local_var_color          => :yellow,
+      :pry_var_color            => :default,         # e.g. _, _pry_, _file_
 
       :instance_var_color       => :blue,        # e.g. @foo
       :class_var_color          => :bright_blue, # e.g. @@foo
@@ -273,9 +355,10 @@ class Pry
       :builtin_global_color     => :cyan,        # e.g. $stdin, $-w, $PID
       :pseudo_global_color      => :cyan,        # e.g. $~, $1..$9, $LAST_MATCH_INFO
 
-      :constant_color           => :default,       # e.g. VERSION, ARGF
+      :constant_color           => :default,     # e.g. VERSION, ARGF
       :class_constant_color     => :blue,        # e.g. Object, Kernel
       :exception_constant_color => :magenta,     # e.g. Exception, RuntimeError
+      :unloaded_constant_color  => :yellow,      # Any constant that is still in .autoload? state
 
       # What should separate items listed by ls? (TODO: we should allow a columnar layout)
       :separator                => "  ",
@@ -300,7 +383,6 @@ class Pry
     self.current_line = 1
     self.line_buffer = [""]
     self.eval_path = "(pry)"
-    self.active_sessions = 0
 
     fix_coderay_colors
   end
@@ -308,19 +390,11 @@ class Pry
   # To avoid mass-confusion, we change the default colour of "white" to
   # "blue" enabling global legibility
   def self.fix_coderay_colors
-      to_fix = if (CodeRay::Encoders::Terminal::TOKEN_COLORS rescue nil)
-                 # CodeRay 1.0.0
-                 CodeRay::Encoders::Terminal::TOKEN_COLORS
-               else
-                 # CodeRay 0.9
-                 begin
-                   require 'coderay/encoders/term'
-                   CodeRay::Encoders::Term::TOKEN_COLORS
-                 rescue => e
-                 end
-               end
+    colors = CodeRay::Encoders::Terminal::TOKEN_COLORS
 
-      to_fix[:comment] = "0;34" if to_fix
+    if colors[:comment].is_a?(String)
+      colors[:comment] = "0;34"
+    end
   end
 
   # Basic initialization.
@@ -339,7 +413,7 @@ class Pry
   # @param [Object] target The object to get a `Binding` object for.
   # @return [Binding] The `Binding` object.
   def self.binding_for(target)
-    if target.is_a?(Binding)
+    if Binding === target
       target
     else
       if TOPLEVEL_BINDING.eval('self') == target
@@ -349,6 +423,34 @@ class Pry
       end
     end
   end
+
+  def self.toplevel_binding
+    unless @toplevel_binding
+      # Grab a copy of the TOPLEVEL_BINDING without any local variables.
+      # This binding has a default definee of Object, and new methods are
+      # private (just as in TOPLEVEL_BINDING).
+      TOPLEVEL_BINDING.eval <<-RUBY
+        def self.__pry__
+          binding
+        end
+        Pry.toplevel_binding = __pry__
+        class << self; undef __pry__; end
+      RUBY
+    end
+    @toplevel_binding.eval('private')
+    @toplevel_binding
+  end
+
+  def self.in_critical_section?
+    @critical_section.to_i > 0
+  end
+
+  def self.critical_section(&block)
+    @critical_section = @critical_section.to_i + 1
+    yield
+  ensure
+    @critical_section -= 1
+  end
 end
 
 Pry.init
