@@ -1,5 +1,26 @@
 require "pry/indent"
 
+##
+# Pry is a powerful alternative to the standard IRB shell for Ruby. It
+# features syntax highlighting, a flexible plugin architecture, runtime
+# invocation and source and documentation browsing.
+#
+# Pry can be started similar to other command line utilities by simply running
+# the following command:
+#
+#     pry
+#
+# Once inside Pry you can invoke the help message:
+#
+#     help
+#
+# This will show a list of available commands and their usage. For more
+# information about Pry you can refer to the following resources:
+#
+# * http://pry.github.com/
+# * https://github.com/pry/pry
+# * the IRC channel, which is #pry on the Freenode network
+#
 class Pry
 
   attr_accessor :input
@@ -8,6 +29,8 @@ class Pry
   attr_accessor :print
   attr_accessor :exception_handler
   attr_accessor :input_stack
+  attr_accessor :quiet
+  alias :quiet? :quiet
 
   attr_accessor :custom_completions
 
@@ -24,6 +47,13 @@ class Pry
 
   attr_accessor :backtrace
 
+  attr_accessor :extra_sticky_locals
+
+  attr_accessor :suppress_output
+
+  # This is exposed via Pry::Command#state.
+  attr_reader :command_state
+
   # Special treatment for hooks as we want to alert people of the
   # changed API
   attr_reader :hooks
@@ -48,12 +78,14 @@ class Pry
   # @option options [Hash] :hooks The defined hook Procs
   # @option options [Array<Proc>] :prompt The array of Procs to use for the prompts.
   # @option options [Proc] :print The Proc to use for the 'print'
+  # @option options [Boolean] :quiet If true, omit the whereami banner when starting.
   #   component of the REPL. (see print.rb)
   def initialize(options={})
     refresh(options)
 
-    @binding_stack     = []
-    @indent            = Pry::Indent.new
+    @binding_stack = []
+    @indent        = Pry::Indent.new
+    @command_state = {}
   end
 
   # Refresh the Pry instance settings from the Pry class.
@@ -63,15 +95,17 @@ class Pry
   def refresh(options={})
     defaults   = {}
     attributes = [
-                   :input, :output, :commands, :print,
+                   :input, :output, :commands, :print, :quiet,
                    :exception_handler, :hooks, :custom_completions,
-                   :prompt, :memory_size, :input_stack
+                   :prompt, :memory_size, :extra_sticky_locals
                  ]
 
     attributes.each do |attribute|
       defaults[attribute] = Pry.send attribute
     end
 
+    defaults[:input_stack] = Pry.input_stack.dup
+
     defaults.merge!(options).each do |key, value|
       send("#{key}=", value) if respond_to?("#{key}=")
     end
@@ -79,6 +113,12 @@ class Pry
     true
   end
 
+  # The currently active `Binding`.
+  # @return [Binding] The currently active `Binding` for the session.
+  def current_context
+    binding_stack.last
+  end
+
   # The current prompt.
   # This is the prompt at the top of the prompt stack.
   #
@@ -105,10 +145,10 @@ class Pry
   # @param [Binding] b The binding to set the local on.
   # @return [Object] The value the local was set to.
   def inject_local(name, value, b)
-    Thread.current[:__pry_local__] = value
-    b.eval("#{name} = Thread.current[:__pry_local__]")
+    Pry.current[:pry_local] = value.is_a?(Proc) ? value.call : value
+    b.eval("#{name} = ::Pry.current[:pry_local]")
   ensure
-    Thread.current[:__pry_local__] = nil
+    Pry.current[:pry_local] = nil
   end
 
   # @return [Integer] The maximum amount of objects remembered by the inp and
@@ -122,48 +162,45 @@ class Pry
     @output_array = Pry::HistoryArray.new(size)
   end
 
-  # Make sure special locals exist at start of session
-  def initialize_special_locals(target)
-    inject_local("_in_", @input_array, target)
-    inject_local("_out_", @output_array, target)
-    inject_local("_pry_", self, target)
-    inject_local("_ex_", last_exception, target)
-    inject_local("_file_", last_file, target)
-    inject_local("_dir_", last_dir, target)
-
-    # without this line we get 1 test failure, ask Mon_Ouie
-    set_last_result(nil, target)
-    inject_local("_", nil, target)
-  end
-  private :initialize_special_locals
-
-  def inject_special_locals(target)
-    special_locals.each_pair do |name, value|
+  # Inject all the sticky locals into the `target` binding.
+  # @param [Binding] target
+  def inject_sticky_locals(target)
+    sticky_locals.each_pair do |name, value|
       inject_local(name, value, target)
     end
   end
 
-  def special_locals
-    {
-      :_in_   => @input_array,
-      :_out_  => @output_array,
+  # Add a sticky local to this Pry instance.
+  # A sticky local is a local that persists between all bindings in a session.
+  # @param [Symbol] name The name of the sticky local.
+  # @yield The block that defines the content of the local. The local
+  #   will be refreshed at each tick of the repl loop.
+  def add_sticky_local(name, &block)
+    sticky_locals[name] = block
+  end
+
+  # @return [Hash] The currently defined sticky locals.
+  def sticky_locals
+    @sticky_locals ||= {
+      :_in_   => proc { @input_array },
+      :_out_  => proc { @output_array },
       :_pry_  => self,
-      :_ex_   => last_exception,
-      :_file_ => last_file,
-      :_dir_  => last_dir,
-      :_      => last_result
-    }
+      :_ex_   => proc { last_exception },
+      :_file_ => proc { last_file },
+      :_dir_  => proc { last_dir },
+      :_      => proc { last_result },
+      :__     => proc { @output_array[-2] }
+    }.merge(extra_sticky_locals)
   end
 
   # Initialize the repl session.
   # @param [Binding] target The target binding for the session.
   def repl_prologue(target)
     exec_hook :before_session, output, target, self
-    initialize_special_locals(target)
+    set_last_result(nil, target)
 
     @input_array << nil # add empty input so _in_ and _out_ match
 
-    Pry.active_sessions += 1
     binding_stack.push target
   end
 
@@ -172,9 +209,8 @@ class Pry
   def repl_epilogue(target)
     exec_hook :after_session, output, target, self
 
-    Pry.active_sessions -= 1
     binding_stack.pop
-    Pry.save_history if Pry.config.history.should_save && Pry.active_sessions == 0
+    Pry.save_history if Pry.config.history.should_save
   end
 
   # Start a read-eval-print-loop.
@@ -187,17 +223,23 @@ class Pry
   #   Pry.new.repl(Object.new)
   def repl(target=TOPLEVEL_BINDING)
     target = Pry.binding_for(target)
-    target_self = target.eval('self')
 
     repl_prologue(target)
 
-    break_data = catch(:breakout) do
-      loop do
-        rep(binding_stack.last)
+    break_data = nil
+    exception = catch(:raise_up) do
+      break_data = catch(:breakout) do
+        loop do
+          throw(:breakout) if binding_stack.empty?
+          rep(binding_stack.last)
+        end
       end
+      exception = false
     end
 
-    break_data || nil
+    raise exception if exception
+
+    break_data
   ensure
     repl_epilogue(target)
   end
@@ -211,7 +253,9 @@ class Pry
     target = Pry.binding_for(target)
     result = re(target)
 
-    show_result(result) if should_print?
+    Pry.critical_section do
+      show_result(result)
+    end
   end
 
   # Perform a read-eval
@@ -225,33 +269,16 @@ class Pry
   def re(target=TOPLEVEL_BINDING)
     target = Pry.binding_for(target)
 
-    compl = Pry::InputCompleter.build_completion_proc(target,
-                                                      instance_eval(&custom_completions))
-
-    if defined? Coolline and input.is_a? Coolline
-      input.completion_proc = proc do |cool|
-        compl.call cool.completed_word
-      end
-    elsif input.respond_to? :completion_proc=
-      input.completion_proc = compl
-    end
-
     # It's not actually redundant to inject them continually as we may have
-    # moved into the scope of a new Binding (e.g the user typed `cd`)
-    inject_special_locals(target)
+    # moved into the scope of a new Binding (e.g the user typed `cd`).
+    inject_sticky_locals(target)
 
     code = r(target)
 
-    result = target.eval(code, Pry.eval_path, Pry.current_line)
-    set_last_result(result, target, code)
-
-    result
+    evaluate_ruby(code, target)
   rescue RescuableException => e
     self.last_exception = e
     e
-  ensure
-    update_input_history(code)
-    exec_hook :after_eval, result, self
   end
 
   # Perform a read.
@@ -268,35 +295,51 @@ class Pry
     target = Pry.binding_for(target)
     @suppress_output = false
 
-    val = ""
     loop do
       begin
         # eval_string will probably be mutated by this method
         retrieve_line(eval_string, target)
-      rescue CommandError, Slop::InvalidOptionError => e
+      rescue CommandError, Slop::InvalidOptionError, MethodSource::SourceNotFoundError => e
+        Pry.last_internal_error = e
         output.puts "Error: #{e.message}"
       end
 
       begin
-        break if complete_expression?(eval_string)
+        break if Pry::Code.complete_expression?(eval_string)
       rescue SyntaxError => e
-        output.puts "SyntaxError: #{e.message.sub(/.*syntax error, */m, '')}"
+        exception_handler.call(output, e.extend(UserError), self)
         eval_string = ""
       end
     end
 
-    @suppress_output = true if eval_string =~ /;\Z/ || eval_string.empty?
+    if eval_string =~ /;\Z/ || eval_string.empty? || eval_string =~ /\A *#.*\n\z/
+      @suppress_output = true
+    end
 
     exec_hook :after_read, eval_string, self
     eval_string
   end
 
+  def evaluate_ruby(code, target = binding_stack.last)
+    target = Pry.binding_for(target)
+    inject_sticky_locals(target)
+    exec_hook :before_eval, code, self
+
+    result = target.eval(code, Pry.eval_path, Pry.current_line)
+    set_last_result(result, target, code)
+  ensure
+    update_input_history(code)
+    exec_hook :after_eval, result, self
+  end
+
   # Output the result or pass to an exception handler (if result is an exception).
   def show_result(result)
     if last_result_is_exception?
-      exception_handler.call output, result, self
+      exception_handler.call(output, result, self)
+    elsif should_print?
+      print.call(output, result)
     else
-      print.call output, result
+      # nothin'
     end
   rescue RescuableException => e
     # Being uber-paranoid here, given that this exception arose because we couldn't
@@ -330,9 +373,24 @@ class Pry
     @indent.reset if eval_string.empty?
 
     current_prompt = select_prompt(eval_string, target)
-    indentation = Pry.config.auto_indent ? @indent.indent_level : ''
+    completion_proc = Pry.config.completer.build_completion_proc(target, self,
+                                                        instance_eval(&custom_completions))
 
-    val = readline("#{current_prompt}#{indentation}")
+    safe_completion_proc = proc{ |*a| Pry.critical_section{ completion_proc.call(*a) } }
+
+    indentation = Pry.config.auto_indent ? @indent.current_prefix : ''
+
+    begin
+      val = readline("#{current_prompt}#{indentation}", safe_completion_proc)
+
+    # Handle <Ctrl+C> like Bash, empty the current input buffer but do not quit.
+    # This is only for ruby-1.9; other versions of ruby do not let you send Interrupt
+    # from within Readline.
+    rescue Interrupt
+      output.puts ""
+      eval_string.replace("")
+      return
+    end
 
     # invoke handler if we receive EOF character (^D)
     if !val
@@ -352,20 +410,26 @@ class Pry
       original_val = "#{indentation}#{val}"
       indented_val = @indent.indent(val)
 
-      if original_val != indented_val && output.tty? && Pry::Helpers::BaseHelpers.use_ansi_codes? && Pry.config.correct_indent
-        output.print @indent.correct_indentation(current_prompt + indented_val, original_val.length - indented_val.length)
+      if output.tty? && Pry::Helpers::BaseHelpers.use_ansi_codes? && Pry.config.correct_indent
+        output.print @indent.correct_indentation(current_prompt, indented_val, original_val.length - indented_val.length)
         output.flush
       end
     else
       indented_val = val
     end
 
+    # Check this before processing the line, because a command might change
+    # Pry's input.
+    interactive = !input.is_a?(StringIO)
+
     begin
       if !process_command(val, eval_string, target)
-        eval_string << "#{indented_val.rstrip}\n" unless val.empty?
+        unless [val, eval_string].all?(&:empty?)
+          eval_string << "#{indented_val.chomp}\n"
+        end
       end
     ensure
-      Pry.history << indented_val unless input.is_a?(StringIO)
+      Pry.history << indented_val if interactive
     end
   end
 
@@ -376,7 +440,8 @@ class Pry
   # @param [String] eval_string The cumulative lines of input.
   # @param [Binding] target The target of the Pry session.
   # @return [Boolean] `true` if `val` is a command, `false` otherwise
-  def process_command(val, eval_string, target)
+  def process_command(val, eval_string = '', target = binding_stack.last)
+    val = val.chomp
     result = commands.process_line(val, {
       :target => target,
       :output => output,
@@ -385,7 +450,7 @@ class Pry
     })
 
     # set a temporary (just so we can inject the value we want into eval_string)
-    Thread.current[:__pry_cmd_result__] = result
+    Pry.current[:pry_cmd_result] = result
 
     # note that `result` wraps the result of command processing; if a
     # command was matched and invoked then `result.command?` returns true,
@@ -395,7 +460,7 @@ class Pry
         # the command that was invoked was non-void (had a return value) and so we make
         # the value of the current expression equal to the return value
         # of the command.
-        eval_string.replace "Thread.current[:__pry_cmd_result__].retval\n"
+        eval_string.replace "::Pry.current[:pry_cmd_result].retval\n"
       end
       true
     else
@@ -496,6 +561,7 @@ class Pry
   # Manage switching of input objects on encountering EOFErrors
   def handle_read_errors
     should_retry = true
+    exception_count = 0
     begin
       yield
     rescue EOFError
@@ -509,19 +575,54 @@ class Pry
       else
         self.input = input_stack.pop
       end
+
       retry
+
+    # Interrupts are handled in r() because they need to tweak eval_string
+    # TODO: Refactor this baby.
+    rescue Interrupt
+      raise
+
+    # If we get a random error when trying to read a line we don't want to automatically
+    # retry, as the user will see a lot of error messages scroll past and be unable to do
+    # anything about it.
+    rescue RescuableException => e
+      puts "Error: #{e.message}"
+      output.puts e.backtrace
+      exception_count += 1
+      if exception_count < 5
+        retry
+      end
+      puts "FATAL: Pry failed to get user input using `#{input}`."
+      puts "To fix this you may be able to pass input and output file descriptors to pry directly. e.g."
+      puts "  Pry.config.input = STDIN"
+      puts "  Pry.config.output = STDOUT"
+      puts "  binding.pry"
+      throw(:breakout)
     end
   end
-
   private :handle_read_errors
 
   # Returns the next line of input to be used by the pry instance.
   # This method should not need to be invoked directly.
   # @param [String] current_prompt The prompt to use for input.
   # @return [String] The next line of input.
-  def readline(current_prompt="> ")
+  def readline(current_prompt="> ", completion_proc=nil)
     handle_read_errors do
-      if input == Readline
+
+      if defined? Coolline and input.is_a? Coolline
+        input.completion_proc = proc do |cool|
+          completions = completion_proc.call cool.completed_word
+          completions.compact
+        end
+      elsif input.respond_to? :completion_proc=
+        input.completion_proc = completion_proc
+      end
+
+      if defined?(Readline) && input == Readline
+        if !$stdout.tty? && $stdin.tty? && !Pry::Helpers::BaseHelpers.windows?
+          Readline.output = File.open('/dev/tty', 'w')
+        end
         input.readline(current_prompt, false) # false since we'll add it manually
       elsif defined? Coolline and input.is_a? Coolline
         input.readline(current_prompt)
@@ -536,11 +637,10 @@ class Pry
   end
 
   # Whether the print proc should be invoked.
-  # Currently only invoked if the output is not suppressed OR the last result
-  # is an exception regardless of suppression.
+  # Currently only invoked if the output is not suppressed.
   # @return [Boolean] Whether the print proc should be invoked.
   def should_print?
-    !@suppress_output || last_result_is_exception?
+    !@suppress_output
   end
 
   # Returns the appropriate prompt to use.
@@ -551,16 +651,43 @@ class Pry
   def select_prompt(eval_string, target)
     target_self = target.eval('self')
 
-    # If input buffer is empty then use normal prompt
-    if eval_string.empty?
-      Array(prompt).first.call(target_self, binding_stack.size - 1, self)
+    open_token = @indent.open_delimiters.any? ? @indent.open_delimiters.last :
+      @indent.stack.last
 
-    # Otherwise use the wait prompt (indicating multi-line expression)
-    else
-      Array(prompt).last.call(target_self, binding_stack.size - 1, self)
+    c = OpenStruct.new(
+                       :object         => target_self,
+                       :nesting_level  => binding_stack.size - 1,
+                       :open_token     => open_token,
+                       :session_line   => Pry.history.session_line_count + 1,
+                       :history_line   => Pry.history.history_line_count + 1,
+                       :expr_number    => input_array.count,
+                       :_pry_          => self,
+                       :binding_stack  => binding_stack,
+                       :input_array    => input_array,
+                       :eval_string    => eval_string,
+                       :cont           => !eval_string.empty?)
+
+    Pry.critical_section do
+      # If input buffer is empty then use normal prompt
+      if eval_string.empty?
+        generate_prompt(Array(prompt).first, c)
+
+      # Otherwise use the wait prompt (indicating multi-line expression)
+      else
+        generate_prompt(Array(prompt).last, c)
+      end
     end
   end
 
+  def generate_prompt(prompt_proc, conf)
+    if prompt_proc.arity == 1
+      prompt_proc.call(conf)
+    else
+      prompt_proc.call(conf.object, conf.nesting_level, conf._pry_)
+    end
+  end
+  private :generate_prompt
+
   # the array that the prompt stack is stored in
   def prompt_stack
     @prompt_stack ||= Array.new
@@ -594,50 +721,45 @@ class Pry
     prompt_stack.size > 1 ? prompt_stack.pop : prompt
   end
 
-  # Determine if a string of code is a complete Ruby expression.
-  # @param [String] code The code to validate.
-  # @return [Boolean] Whether or not the code is a complete Ruby expression.
-  # @raise [SyntaxError] Any SyntaxError that does not represent incompleteness.
-  # @example
-  #   complete_expression?("class Hello") #=> false
-  #   complete_expression?("class Hello; end") #=> true
-  def complete_expression?(str)
-    if defined?(Rubinius::Melbourne19) && RUBY_VERSION =~ /^1\.9/
-      Rubinius::Melbourne19.parse_string(str, Pry.eval_path)
-    elsif defined?(Rubinius::Melbourne)
-      Rubinius::Melbourne.parse_string(str, Pry.eval_path)
-    else
-      catch(:valid) do
-        Helpers::BaseHelpers.silence_warnings do
-          eval("BEGIN{throw :valid}\n#{str}", binding, Pry.eval_path)
-        end
-      end
-    end
+  # Raise an exception out of Pry.
+  #
+  # See Kernel#raise for documentation of parameters.
+  # See rb_make_exception for the inbuilt implementation.
+  #
+  # This is necessary so that the raise-up command can tell the
+  # difference between an exception the user has decided to raise,
+  # and a mistake in specifying that exception.
+  #
+  # (i.e. raise-up RunThymeError.new should not be the same as
+  #  raise-up NameError, "unititialized constant RunThymeError")
+  #
+  def raise_up_common(force, *args)
+    exception = if args == []
+                  last_exception || RuntimeError.new
+                elsif args.length == 1 && args.first.is_a?(String)
+                  RuntimeError.new(args.first)
+                elsif args.length > 3
+                  raise ArgumentError, "wrong number of arguments"
+                elsif !args.first.respond_to?(:exception)
+                  raise TypeError, "exception class/object expected"
+                elsif args.length === 1
+                  args.first.exception
+                else
+                  args.first.exception(args[1])
+                end
 
-    # Assert that a line which ends with a , is incomplete.
-    str !~ /[,]\z/
-  rescue SyntaxError => e
-    if incomplete_user_input_exception?(e)
-      false
+    raise TypeError, "exception object expected" unless exception.is_a? Exception
+
+    exception.set_backtrace(args.length === 3 ? args[2] : caller(1))
+
+    if force || binding_stack.one?
+      binding_stack.clear
+      throw :raise_up, exception
     else
-      raise e
-    end
-  end
-
-  # Check whether the exception indicates that the user should input more.
-  #
-  # @param [SyntaxError] the exception object that was raised.
-  # @param [Array<String>] The stack frame of the function that executed eval.
-  # @return [Boolean]
-  #
-  def incomplete_user_input_exception?(ex)
-    case ex.message
-    when /unexpected (\$end|end-of-file|END_OF_FILE)/, # mri, jruby, ironruby
-        /unterminated (quoted string|string|regexp) meets end of file/, # "quoted string" is ironruby
-        /missing 'end' for/, /: expecting '[})\]]'$/, /can't find string ".*" anywhere before EOF/, /expecting keyword_end/ # rbx
-      true
-    else
-      false
+      binding_stack.pop
+      raise exception
     end
   end
+  def raise_up(*args); raise_up_common(false, *args); end
+  def raise_up!(*args); raise_up_common(true, *args); end
 end
